import logging
from datetime import datetime as dt
from copy import deepcopy as copy

logger = logging.getLogger(__name__)


class DKANDatasetFactory(object):
    """
    docstring
    """

    def __init__(
        self,
        default_dataset_metadata=None,
        auto_set_modified_time=True,
        modified_time_stamp=None,
    ):
        self.auto_set_modified_time = auto_set_modified_time

        self.default_dataset_metadata = default_dataset_metadata
        self.dkan_time_format = "%Y-%m-%dT%H:%M:%S"

        if self.auto_set_modified_time:
            self.modified_time_stamp = dt.now().strftime(self.dkan_time_format)
        else:
            self.modified_time_stamp = modified_time_stamp

    def _get_distribution_format_from_url(self, download_url):
        assert download_url.endswith(".csv"), "unsupported file format"
        return {"mediaType": "text/csv", "format": "csv"}

    def create_dataset_body_with_one_distribution(
        self,
        dataset_title,
        distribution_title,
        distribution_description,
        download_url,
        identifier=None,
    ):
        dataset_params = copy(self.default_dataset_metadata)
        dataset_param_updates = {
            "title": dataset_title,
        }
        if identifier is not None:
            dataset_param_updates["identifier"] = identifier

        if self.auto_set_modified_time:
            dataset_param_updates["modified"] = self.modified_time_stamp

        dataset_params.update(dataset_param_updates)

        distribution_params = {
            "title": distribution_title,
            "description": distribution_description,
            "downloadURL": download_url,
        }

        format = self._get_distribution_format_from_url(download_url)
        distribution_params.update(format)

        if "distribution" not in dataset_params:
            dataset_params["distribution"] = []

        dataset_params["distribution"].append(distribution_params)

        return dataset_params



# ignore redefining from outer scope, false positive from fixtures
# pylint: disable=W0621
import pytest
from dkany.client import DKANClient


# ----------------------------------
# Fixtures
# ----------------------------------
@pytest.fixture(scope="function")
def dkan_client():
    client = DKANClient(base_url="https://edit.data.medicaid.gov")
    return client


@pytest.fixture(scope="function")
def read_dataset_id():
    return "9e407144-9ed9-5cee-937a-17d65b07a9a7"


# ----------------------------------
# Tests
# ----------------------------------
def test_search(dkan_client):
    title_success = "Product Data for Newly Reported Drugs in the Medicaid Drug Rebate Program 2023-02-06-to-2023-02-12"
    search_success = dkan_client.search(title=title_success)
    assert len(search_success) == 1
    print("Search success:", len(search_success))

    title_fail = "Product Data for asdfasdf"
    search_fail = dkan_client.search(title=title_fail)
    assert len(search_fail) == 0
    print("Search fail:", len(search_fail))


def test_check_dataset_exists(dkan_client, read_dataset_id):
    exists = dkan_client.check_dataset_exists(read_dataset_id)
    print(f"dataset {exists} exists")
    assert exists


def test_get_dataset_metadata(dkan_client, read_dataset_id):
    metadata = dkan_client.get_dataset_metadata(read_dataset_id)
    assert metadata["identifier"] == read_dataset_id


def test_get_data_by_dataset_identifier(dkan_client, read_dataset_id):
    results = dkan_client.get_data_by_dataset_identifier(read_dataset_id)
    assert len(results) > 0



import logging
import json
from json import JSONEncoder

import luigi
from luigi.freezing import FrozenOrderedDict

from dkany.client.client import DKANClient

logger = logging.getLogger(__name__)


class DkanTarget(luigi.Target):
    def __init__(
        self,
        dataset_identifier=None,
        search_params=None,
        search_results_filter=None,
        datastore_idx=0,
        dkan_client_params=None,
        dkan_client=None,
    ):
        if (search_params is None) and (dataset_identifier is None):
            raise ValueError("Either dataset_identifier or search_params must be specified")

        if dkan_client is None:
            params = dkan_client_params or {}
            self.client = DKANClient(**params)
        else:
            self.client = dkan_client

        self.search_params = search_params
        self.search_results_filter = search_results_filter
        self.dataset_identifier = dataset_identifier
        self.datastore_idx = datastore_idx

    def verify_current_check(self, _dataset_metadata):
        # This check accomplishes nothing,
        # meant to be specialized in a subclass
        return True

    def exists(self):
        logger.debug("Checking datasets existance")
        dataset_identifier = None
        if self.dataset_identifier is not None:
            if self.client.check_dataset_exists(self.dataset_identifier):
                dataset_identifier = self.dataset_identifier
                dataset_found = True
            else:
                dataset_found = False

        elif self.search_params is not None:
            raw_results = self.client.search(**self.search_params)
            results = self.client.filter_search_results(raw_results, self.search_results_filter)
            n_results = len(results.keys())

            logger.debug("n_results %s", n_results)

            if n_results == 1:
                dataset_identifier = results[list(results.keys())[0]]["identifier"]
                dataset_found = True
            elif n_results == 0:
                dataset_found = False
            elif n_results > 1:
                error_text = (
                    f"{n_results} results was returned, when exactly 1 or 0 is required. " +
                    "The search criteria may need to be more specific, or " +
                    "a search_results_filter may need to be applied. " +
                    f"Search Params: \n{self.search_params}\n " +
                    f"search_results_filter: \n{self.search_results_filter}\n " +
                    f"Search Results: \n{results}\n"
                )
                raise ValueError(error_text)
            else:
                raise ValueError(f"n_results of {n_results} unexpected ")
        else:
            raise (
                ValueError(
                    "unexpectedly Both dataset Identifier and search results appear to be none"
                )
            )

        if dataset_found:
            dataset_metadata = self.client.get_dataset_metadata(dataset_identifier)
            current = self.verify_current_check(dataset_metadata)
        else:
            current = False

        if not dataset_found:
            logger.debug("Dataset not found, promting a creation")
        elif dataset_found and (not current):
            logger.debug("Dataset found but not current, promiting an update")
        elif dataset_found and (current):
            logger.debug("Dataset found and current, no action will be taken")
        else:
            raise ValueError("Unexpected State")

        return current


class FrozenOrderedDictEncoder(JSONEncoder):
    def default(self, o):
        if isinstance(o, FrozenOrderedDict):
            return o.get_wrapped()
        return json.JSONEncoder.default(self, o)


class WriteToDkanDatasetIdBySearch(luigi.Task):
    dkan_client_params = luigi.DictParameter(
        description="The paramaters you'd like to use to make the dkan client"
    )
    dataset_search_params = luigi.DictParameter(
        description="The search paramaters you'd like to use to uniqely identify your dataset"
    )
    dataset_search_results_filter = luigi.DictParameter(
        description="The search paramaters you'd like to use to uniqely identify your dataset",
        default={},
    )
    dataset_request_body = luigi.DictParameter(
        description="The body of the request that you'd like to post"
    )
    datastore_idx = luigi.IntParameter(description="", default=0)

    def make_dkan_client(self):
        return DKANClient(**self.dkan_client_params)  # pylint: disable=E1134

    def output(self):
        target = DkanTarget(
            search_params=self.dataset_search_params,
            search_results_filter=self.dataset_search_results_filter,
            datastore_idx=self.datastore_idx,
            dkan_client=self.make_dkan_client(),
        )

        return target

    def run(self):
        logger.debug("Running CreateOrUpdateDkanDataset")
        dkan_client = self.make_dkan_client()

        raw_results = dkan_client.search(**self.dataset_search_params)  # pylint: disable=E1134
        results = dkan_client.filter_search_results(raw_results, self.dataset_search_results_filter)

        n_results = len(results.keys())

        # This is the easiest method to resursively transform everything to a dict
        request_body = json.loads(
            json.dumps(self.dataset_request_body, cls=FrozenOrderedDictEncoder)
        )

        if n_results == 1:
            dataset_identifier = results[list(results.keys())[0]]["identifier"]
            logger.debug("Updating existing dataset on DKAN. ID %s", dataset_identifier)
            request_body["identifier"] = dataset_identifier
            _ = dkan_client.update_dataset(dataset_identifier, request_body)
        elif n_results == 0:
            logger.debug("Creating New Dataset on DKAN")
            _ = dkan_client.create_dataset(request_body)
        else:
            raise ValueError(f"n_results of {n_results} unexpected ")



from dkany.client import DKANClient as DkanyClient

def main():
    client = DkanyClient(
        base_url = "https://edit.data.medicaid.gov"
    )

    test_dataset_id = "9e407144-9ed9-5cee-937a-17d65b07a9a7"

    exists = client.check_dataset_exists(test_dataset_id)

    print(f"dataset {exists} exits")

if __name__ == "__main__":
    main()


import pysftp


class SftpClient(object):
    def __init__(
        self,
        host=None,
        username=None,
        private_key_file=None,
        private_key_pass=None,
        default_remote_path=None,
        hostkeys=None,
    ):
        cnopts = pysftp.CnOpts()
        # Attempted to get a valid hostkey from Akamai with a waka ticket,
        # but because of details of their setup it seems like it might not be possible
        # to get a relevent one
        cnopts.hostkeys = hostkeys

        self.connection = pysftp.Connection(
            host,
            username,
            private_key=private_key_file,
            private_key_pass=private_key_pass,
            default_path=default_remote_path,
            cnopts=cnopts,
        )



class Error(Exception):
    """Base class for exceptions in this module."""
    message: str

    def __str__(self):
        return self.message

    def __repr__(self):
        return self.message


class BadResponse(Error):
    def __init__(self, response, acceptable_status_codes):
        status_code = response.status_code
        message = []
        message.append(
            "Status code returned not in acceptable status codes for this response"
        )
        message.append(
            f"Returned: {status_code}:{response.reason}, Acceptable Codes {acceptable_status_codes}"
        )
        message.append(response.text)

        self.response = response
        self.message = "\n".join(message)



# import os
# from dkany.data_hosts.sftp import SftpClient

def test_akamai_sftp_client():
    pass
    # akamai_sftp_client = SftpClient(
    #     host = "cmsstorage.upload.akamai.com",
    #     username = "sshacs",
    #     private_key_file = os.environ['DKAN_AKAMAI_PRIVATE_KEY'],
    #     private_key_pass = os.environ['DKAN_AKAMAI_PRIVATE_KEY_PASSWORD'],
    #     default_remote_path='/399963/questions.medicaid.gov/production/wwwroot/data/scorecard/'
    # )

    # cwd = akamai_sftp_client.connection.getcwd()

    # print(cwd)

    # output = akamai_sftp_client.connection.listdir()

    # print(output)
    # raise(ValueError("hi"))
    


import logging
from copy import deepcopy as copy
from datetime import datetime as dt
from typing import List, Optional

import requests
from requests.cookies import RequestsCookieJar
from requests_toolbelt import sessions # type: ignore

from dkany.client.errors import BadResponse

logger = logging.getLogger(__name__)


def url_join(url_part_list):
    return "/".join(url_part_list)


class DKANClient:
    """
    docstring
    """

    def __init__(
        self,
        base_url: Optional[str] = None,
        cookie_dict: Optional[dict] = None,
        user_name: Optional[str] = None,
        password: Optional[str] = None,
    ):
        self.base_url = base_url

        logger.info("Creating DKAN client for %s", self.base_url)

        session = sessions.BaseUrlSession(self.base_url)
        if user_name is not None:
            session.auth = (user_name, password)
            self.user_name = user_name
        else:
            self.user_name = "anonymous"

        if cookie_dict is not None:
            cookies = RequestsCookieJar()
            cookies = requests.utils.add_dict_to_cookiejar(cookies, cookie_dict)
            session.cookies = cookies

        self.session = session

        self.search_url = "api/1/search?_format=json"
        self.post_new_dataset_url = "api/1/metastore/schemas/dataset/items?_format=json"
        self.existing_dataset_url = (
            "api/1/metastore/schemas/dataset/items/{dataset_identifier}?_format=json"
        )
        self.revise_dataset_url = "api/1/metastore/schemas/dataset/items/{dataset_identifier}/revisions?_format=json"
        self.query_datastore_url = (
            "api/1/datastore/query/{dataset_identifier}/{datastore_idx}?_format=json"
        )

        self.hide_dataset_dict = {"state": "hidden", "message": "hiding dataset"}
        self.publish_dataset_dict = {
            "state": "published",
            "message": "publishing dataset",
        }

        self.dkan_time_format = "%Y-%m-%dT%H:%M:%S"

    def __repr__(self) -> str:
        return f"DKAN client for {self.base_url} with user {self.user_name}"

    def __str__(self) -> str:
        return f"DKAN client for {self.base_url} with user {self.user_name}"

    def _process_response(
        self, response, acceptable_responses: Optional[List[int]] = None
    ):
        acceptable_responses = acceptable_responses or [200, 201]
        if response.status_code not in acceptable_responses:
            raise BadResponse(response, acceptable_responses)
        out = response.json()
        return out

    def _paged_search(self, params, page):
        params["page"] = page

        response = self.session.get(self.search_url, params=params)

        return self._process_response(response)

    def _search_all_pages(self, params):
        page = 1
        out = self._paged_search(params, page)
        total = int(out["total"])
        all_results = copy(out["results"])

        if total > 0:
            while len(all_results.keys()) < total:
                page += 1
                out = self._paged_search(params, page)
                all_results.update(out["results"])

        if all_results != []:
            assert len(all_results.keys()) == total

        return all_results

    def search(
        self, title: Optional[str] = None, tags=None, categories=None, page="ALL"
    ):
        params = {}
        if title is not None:
            params["title"] = title
        if tags is not None:
            params["keyword"] = tags
        if categories is not None:
            params["theme"] = categories

        # Paging Logic
        if page != "ALL":
            out = self._paged_search(params, page)["results"]
        else:
            out = self._search_all_pages(params)

        # search returns a dict if it finds something
        # and and empty list if it does not find anything
        # this is for type consistentcy
        if out == []:
            out = {}

        return out

    def filter_search_results(self, search_results, filter_params):
        if filter_params is None:
            return search_results
        if len(filter_params.keys()) == 0:
            return search_results

        inital_search_results = list(search_results.items())

        for search_key, search_result_value in inital_search_results:
            for filter_key, filter_value in filter_params.items():
                if search_result_value[filter_key] != filter_value:
                    search_results.pop(search_key)
                    break

        return search_results

    def create_dataset(self, body):
        response = self.session.post(self.post_new_dataset_url, json=body)
        return self._process_response(response)

    def delete_dataset(self, dataset_identifier):
        response = self.session.delete(
            self.existing_dataset_url.format(dataset_identifier=dataset_identifier)
        )
        return self._process_response(response)

    def update_dataset(self, dataset_identifier, body):
        response = self.session.put(
            self.existing_dataset_url.format(dataset_identifier=dataset_identifier),
            json=body,
        )
        return self._process_response(response)

    def mark_dataset_hidden(self, dataset_identifier, message=""):
        """
        Sets dataset accesslevel to "hidden"
        Hides dataset from searches made on data.medicare.gov user interface
        """
        if message:
            self.hide_dataset_dict["message"] = message
        response = self.session.post(
            self.revise_dataset_url.format(dataset_identifier=dataset_identifier),
            json=self.hide_dataset_dict,
        )
        return self._process_response(response)

    def mark_dataset_public(self, dataset_identifier, message=""):
        """
        Sets dataset accesslevel to "published"
        Makes a dataset searchable through data.medicare.gov user interface
        """
        if message:
            self.publish_dataset_dict["message"] = message
        response = self.session.post(
            self.revise_dataset_url.format(dataset_identifier=dataset_identifier),
            json=self.publish_dataset_dict,
        )
        return self._process_response(response)

    def get_dataset_metadata(self, dataset_identifier):
        response = self.session.get(
            self.existing_dataset_url.format(dataset_identifier=dataset_identifier),
            params={"_format": "json"},
        )
        return self._process_response(response)

    def check_dataset_exists(self, dataset_identifier):
        try:
            _ = self.get_dataset_metadata(dataset_identifier)
            return True
        except BadResponse:
            return False

    def trigger_dataset_reimport(self, dataset_identifier):
        body = self.get_dataset_metadata(dataset_identifier)
        body["modified"] = dt.now().strftime(self.dkan_time_format)
        return self.update_dataset(dataset_identifier, body)

    def get_full_query_url(self, dataset_identifier, datastore_idx=0):
        return url_join(
            [
                self.base_url,
                self.query_datastore_url.format(
                    dataset_identifier=dataset_identifier, datastore_idx=datastore_idx
                ),
            ]
        )

    def get_data_by_dataset_identifier(self, dataset_identifier, datastore_idx=0):
        response = self.session.get(
            self.query_datastore_url.format(
                dataset_identifier=dataset_identifier, datastore_idx=datastore_idx
            )
        )
        return self._process_response(response)



import logging
import os

import yaml

logger = logging.getLogger(__name__)


def get_project_root(directories_to_go_up=3):
    project_dir = os.path.abspath(__file__)

    for _ in range(directories_to_go_up):
        project_dir = os.path.dirname(project_dir)

    return project_dir


def read_yaml_file(full_file_path):
    with open(full_file_path, "r", encoding="utf-8") as stream:
        out_dict = yaml.safe_load(stream)
    return out_dict
